@@ -27,6 +27,7 @@
 #include <sys/types.h>
 #include <errno.h>
 #include <limits.h>
+#include <stdio.h>
 
 #include <ell/ell.h>
 
@@ -41,6 +42,8 @@
 #include "src/agent.h"
 #include "src/device.h"
 #include "src/wiphy.h"
+#include "src/eap.h"
+#include "src/knownnetworks.h"
 #include "src/network.h"
 
 struct network {
@@ -48,28 +51,25 @@ struct network {
 	struct device *device;
 	struct network_info *info;
 	unsigned char *psk;
+	char *passphrase;
 	unsigned int agent_request;
 	struct l_queue *bss_list;
 	struct l_settings *settings;
+	struct l_queue *secrets;
 	bool update_psk:1;  /* Whether PSK should be written to storage */
 	bool ask_psk:1; /* Whether we should force-ask agent for PSK */
 	int rank;
 };
 
-static struct l_queue *networks = NULL;
+static struct l_queue *networks;
 
 static bool network_settings_load(struct network *network)
 {
-	const char *strtype;
-
 	if (network->settings)
 		return true;
 
-	strtype = security_to_str(network_get_security(network));
-	if (!strtype)
-		return false;
-
-	network->settings = storage_network_open(strtype, network->info->ssid);
+	network->settings = storage_network_open(network_get_security(network),
+							network->info->ssid);
 
 	return network->settings != NULL;
 }
@@ -82,33 +82,14 @@ static void network_settings_close(struct network *network)
 	l_free(network->psk);
 	network->psk = NULL;
 
+	l_free(network->passphrase);
+	network->passphrase = NULL;
+
 	l_settings_free(network->settings);
 	network->settings = NULL;
 }
 
-static int timespec_compare(const void *a, const void *b, void *user_data)
-{
-	const struct network_info *ni_a = a;
-	const struct network_info *ni_b = b;
-	const struct timespec *tsa = &ni_a->connected_time;
-	const struct timespec *tsb = &ni_b->connected_time;
-
-	if (tsa->tv_sec > tsb->tv_sec)
-		return -1;
-
-	if (tsa->tv_sec < tsb->tv_sec)
-		return 1;
-
-	if (tsa->tv_nsec > tsb->tv_nsec)
-		return -1;
-
-	if (tsa->tv_nsec < tsb->tv_nsec)
-		return -1;
-
-	return 0;
-}
-
-static bool network_info_match(const void *a, const void *b)
+bool network_info_match(const void *a, const void *b)
 {
 	const struct network_info *ni_a = a;
 	const struct network_info *ni_b = b;
@@ -122,32 +103,28 @@ static bool network_info_match(const void *a, const void *b)
 	return true;
 }
 
-bool network_seen(struct network *network, struct timespec *when)
+static bool network_info_ptr_match(const void *a, const void *b)
 {
-	/*
-	 * Update the last seen time.  Note this is not preserved across
-	 * the network going out of range and back, or program restarts.
-	 * It may be desirable for it to be preserved in some way but
-	 * without too frequent filesystem writes.
-	 */
-	memcpy(&network->info->seen_time, when, sizeof(struct timespec));
-
-	return true;
+	return a == b;
 }
 
-bool network_connected(struct network *network)
+static bool network_secret_check_cacheable(void *data, void *user_data)
+{
+	struct eap_secret_info *secret = data;
+
+	return secret->cache_policy == EAP_CACHE_NEVER;
+}
+
+void network_connected(struct network *network)
 {
 	int err;
-	const char *strtype;
 
-	l_queue_remove(networks, network->info);
-	l_queue_push_head(networks, network->info);
-
-	strtype = security_to_str(network_get_security(network));
-	if (!strtype)
-		return false;
-
-	err = storage_network_touch(strtype, network->info->ssid);
+	/*
+	 * This triggers an update to network->info->connected_time and
+	 * other possible actions in knownnetworks.c.
+	 */
+	err = storage_network_touch(network_get_security(network),
+					network->info->ssid);
 	switch (err) {
 	case 0:
 		break;
@@ -161,21 +138,17 @@ bool network_connected(struct network *network)
 		 */
 		network->settings = l_settings_new();
 
-		storage_network_sync(strtype, network->info->ssid,
+		storage_network_sync(network_get_security(network),
+					network->info->ssid,
 					network->settings);
 		break;
 	default:
-		return false;
+		l_error("Error %i touching network config", err);
+		break;
 	}
 
-	err = storage_network_get_mtime(strtype, network->info->ssid,
-					&network->info->connected_time);
-	if (err < 0)
-		return false;
-
-	network->info->is_known = true;
-
-	return true;
+	l_queue_foreach_remove(network->secrets,
+				network_secret_check_cacheable, network);
 }
 
 void network_disconnected(struct network *network)
@@ -183,21 +156,31 @@ void network_disconnected(struct network *network)
 	network_settings_close(network);
 }
 
+struct network_find_rank_data {
+	const struct network_info *info;
+	int n;
+};
+
+static bool network_find_rank_update(const struct network_info *network,
+					void *user_data)
+{
+	struct network_find_rank_data *data = user_data;
+
+	if (network == data->info)
+		return false;
+
+	if (network->seen_count)
+		data->n++;
+
+	return true;
+}
+
 static int network_find_rank_index(const struct network_info *info)
 {
-	const struct l_queue_entry *entry;
-	int n;
+	struct network_find_rank_data data = { info, 0 };
 
-	for (n = 0, entry = l_queue_get_entries(networks); entry;
-						entry = entry->next) {
-		struct network_info *network = entry->data;
-
-		if (network == info)
-			return n;
-
-		if (network->is_known && network->seen_count)
-			n++;
-	}
+	if (!known_networks_foreach(network_find_rank_update, &data))
+		return data.n;
 
 	return -1;
 }
@@ -249,7 +232,7 @@ bool network_rankmod(const struct network *network, double *rankmod)
 	return true;
 }
 
-static void network_info_free(void *data)
+void network_info_free(void *data)
 {
 	struct network_info *network = data;
 
@@ -259,12 +242,18 @@ static void network_info_free(void *data)
 static struct network_info *network_info_get(const char *ssid,
 						enum security security)
 {
-	struct network_info *network, search;
+	struct network_info *network;
 
-	search.type = security;
-	strcpy(search.ssid, ssid);
+	network = known_networks_find(ssid, security);
 
-	network = l_queue_find(networks, network_info_match, &search);
+	if (!network) {
+		struct network_info search;
+
+		search.type = security;
+		strcpy(search.ssid, ssid);
+
+		network = l_queue_find(networks, network_info_match, &search);
+	}
 
 	if (!network) {
 		network = l_new(struct network_info, 1);
@@ -281,17 +270,14 @@ static struct network_info *network_info_get(const char *ssid,
 
 static void network_info_put(struct network_info *network)
 {
-	if (!networks)
-		return;
-
 	if (--network->seen_count)
 		return;
 
-	if (network->is_known)
+	if (!networks)
 		return;
 
-	l_queue_remove(networks, network);
-	network_info_free(network);
+	if (l_queue_remove(networks, network))
+		network_info_free(network);
 }
 
 struct network *network_create(struct device *device, const char *ssid,
@@ -333,11 +319,24 @@ const uint8_t *network_get_psk(const struct network *network)
 	return network->psk;
 }
 
+const char *network_get_passphrase(const struct network *network)
+{
+	return network->passphrase;
+}
+
+struct l_queue *network_get_secrets(const struct network *network)
+{
+	return network->secrets;
+}
+
 bool network_set_psk(struct network *network, const uint8_t *psk)
 {
 	if (network->info->type != SECURITY_PSK)
 		return false;
 
+	if (!network_settings_load(network))
+		return false;
+
 	l_free(network->psk);
 	network->psk = l_memdup(psk, 32);
 	return true;
@@ -355,34 +354,122 @@ struct l_settings *network_get_settings(const struct network *network)
 	return network->settings;
 }
 
+static bool network_set_8021x_secrets(struct network *network)
+{
+	const struct l_queue_entry *entry;
+
+	if (!network->settings)
+		return false;
+
+	for (entry = l_queue_get_entries(network->secrets); entry;
+			entry = entry->next) {
+		struct eap_secret_info *secret = entry->data;
+		char *setting;
+
+		switch (secret->type) {
+		case EAP_SECRET_LOCAL_PKEY_PASSPHRASE:
+		case EAP_SECRET_REMOTE_PASSWORD:
+			if (!l_settings_set_string(network->settings,
+							"Security", secret->id,
+							secret->value))
+				return false;
+			break;
+
+		case EAP_SECRET_REMOTE_USER_PASSWORD:
+			if (!l_settings_set_string(network->settings,
+							"Security", secret->id,
+							secret->value))
+				return false;
+
+			if (secret->id2)
+				setting = secret->id2;
+			else {
+				setting = alloca(strlen(secret->id) + 10);
+				sprintf(setting, "%s-Password", secret->id);
+			}
+
+			if (!l_settings_set_string(network->settings,
+							"Security", setting,
+							secret->value + 1 +
+							strlen(secret->value)))
+				return false;
+
+			break;
+		}
+	}
+
+	return true;
+}
+
+static int network_load_psk(struct network *network)
+{
+	size_t len;
+	const char *psk = l_settings_get_value(network->settings,
+						"Security", "PreSharedKey");
+
+	if (!psk)
+		return -ENOKEY;
+
+	l_free(network->psk);
+	network->psk = l_util_from_hexstring(psk, &len);
+	if (!network->psk)
+		return -EINVAL;
+
+	if (len != 32) {
+		l_free(network->psk);
+		network->psk = NULL;
+		return -EINVAL;
+	}
+
+	return 0;
+}
+
 void network_sync_psk(struct network *network)
 {
 	char *hex;
+	struct l_settings *fs_settings;
 
 	if (!network->update_psk)
 		return;
 
 	network->update_psk = false;
+
+	fs_settings = storage_network_open(SECURITY_PSK, network->info->ssid);
 	hex = l_util_hexstring(network->psk, 32);
 	l_settings_set_value(network->settings, "Security",
 						"PreSharedKey", hex);
+	l_settings_set_value(fs_settings, "Security", "PreSharedKey", hex);
 	l_free(hex);
-	storage_network_sync("psk", network->info->ssid, network->settings);
+
+	storage_network_sync(SECURITY_PSK, network->info->ssid, fs_settings);
+	l_settings_free(fs_settings);
 }
 
 int network_autoconnect(struct network *network, struct scan_bss *bss)
 {
 	struct wiphy *wiphy = device_get_wiphy(network->device);
 	bool is_autoconnectable;
+	bool is_rsn;
+	int ret;
 
 	switch (network_get_security(network)) {
 	case SECURITY_NONE:
+		is_rsn = false;
 		break;
 	case SECURITY_PSK:
-	{
+		if (network->ask_psk)
+			return -ENOKEY;
+
+		/* Fall through */
+	case SECURITY_8021X:
+		is_rsn = true;
+		break;
+	default:
+		return -ENOTSUP;
+	}
+
+	if (is_rsn) {
 		struct ie_rsn_info rsn;
-		const char *psk;
-		size_t len;
 
 		memset(&rsn, 0, sizeof(rsn));
 		scan_bss_get_rsn_info(bss, &rsn);
@@ -392,52 +479,47 @@ int network_autoconnect(struct network *network, struct scan_bss *bss)
 			l_debug("Cipher mis-match");
 			return -ENETUNREACH;
 		}
-
-		if (network->ask_psk)
-			return -ENOKEY;
-
-		if (!network_settings_load(network))
-			return -ENOKEY;
-
-		psk = l_settings_get_value(network->settings, "Security",
-						"PreSharedKey");
-
-		if (!psk) {
-			network_settings_close(network);
-			return -ENOKEY;
-		}
-
-		l_free(network->psk);
-		network->psk = l_util_from_hexstring(psk, &len);
-
-		if (network->psk && len != 32) {
-			network_settings_close(network);
-			return -ENOKEY;
-		}
-
-		break;
-	}
-	case SECURITY_8021X:
-		if (!network_settings_load(network))
-			return -ENOKEY;
-
-		break;
-	default:
-		return -ENOTSUP;
 	}
 
+	if (!network_settings_load(network))
+		return -ENOKEY;
+
+	/* If no entry, default to Autoconnectable=True */
 	if (!l_settings_get_bool(network->settings, "Settings",
 					"Autoconnect", &is_autoconnectable))
-		goto connect;
+		is_autoconnectable = true;
 
-	if (!is_autoconnectable) {
-		network_settings_close(network);
-		return -EPERM;
+	ret = -EPERM;
+	if (!is_autoconnectable)
+		goto close_settings;
+
+	if (network_get_security(network) == SECURITY_PSK) {
+		ret = network_load_psk(network);
+		if (ret < 0)
+			goto close_settings;
+	} else if (network_get_security(network) == SECURITY_8021X) {
+		struct l_queue *missing_secrets = NULL;
+
+		ret = eap_check_settings(network->settings, network->secrets,
+					"EAP-", true, &missing_secrets);
+		if (ret < 0)
+			goto close_settings;
+
+		ret = -ENOKEY;
+		if (!l_queue_isempty(missing_secrets)) {
+			l_queue_destroy(missing_secrets, eap_secret_info_free);
+			goto close_settings;
+		}
+
+		if (!network_set_8021x_secrets(network))
+			goto close_settings;
 	}
 
-connect:
-	device_connect_network(network->device, network, bss, NULL);
-	return 0;
+	return __device_connect_network(network->device, network, bss);
+
+close_settings:
+	network_settings_close(network);
+	return ret;
 }
 
 void network_connect_failed(struct network *network)
@@ -450,6 +532,9 @@ void network_connect_failed(struct network *network)
 		network->update_psk = false;
 		network->ask_psk = true;
 	}
+
+	l_queue_destroy(network->secrets, eap_secret_info_free);
+	network->secrets = NULL;
 }
 
 bool network_bss_add(struct network *network, struct scan_bss *bss)
@@ -552,6 +637,8 @@ static void passphrase_callback(enum agent_result result,
 
 	l_free(network->psk);
 	network->psk = l_malloc(32);
+	l_free(network->passphrase);
+	network->passphrase = l_strdup(passphrase);
 
 	if (crypto_psk_from_passphrase(passphrase,
 					(uint8_t *) network->info->ssid,
@@ -584,33 +671,13 @@ static struct l_dbus_message *network_connect_psk(struct network *network,
 					struct l_dbus_message *message)
 {
 	struct device *device = network->device;
-	const char *psk;
 
 	l_debug("");
 
-	if (network_settings_load(network)) {
-		psk = l_settings_get_value(network->settings, "Security",
-						"PreSharedKey");
-
-		if (psk) {
-			size_t len;
-
-			l_debug("psk: %s", psk);
-
-			l_free(network->psk);
-			network->psk = l_util_from_hexstring(psk, &len);
-
-			l_debug("len: %zd", len);
-
-			if (network->psk && len != 32) {
-				l_debug("Can't parse PSK");
-				l_free(network->psk);
-				network->psk = NULL;
-			}
-		}
-	} else {
+	if (network_settings_load(network))
+		network_load_psk(network);
+	else
 		network->settings = l_settings_new();
-	}
 
 	l_debug("ask_psk: %s", network->ask_psk ? "true" : "false");
 
@@ -630,6 +697,252 @@ static struct l_dbus_message *network_connect_psk(struct network *network,
 	return NULL;
 }
 
+struct eap_secret_request {
+	struct network *network;
+	struct eap_secret_info *secret;
+	struct l_queue *pending_secrets;
+	void (*callback)(enum agent_result result,
+				struct l_dbus_message *message,
+				struct eap_secret_request *req);
+};
+
+static void eap_secret_request_free(void *data)
+{
+	struct eap_secret_request *req = data;
+
+	eap_secret_info_free(req->secret);
+	l_queue_destroy(req->pending_secrets, eap_secret_info_free);
+	l_free(req);
+}
+
+static bool eap_secret_info_match_local(const void *a, const void *b)
+{
+	const struct eap_secret_info *info = a;
+
+	return info->type == EAP_SECRET_LOCAL_PKEY_PASSPHRASE;
+}
+
+static void eap_password_callback(enum agent_result result, const char *value,
+					struct l_dbus_message *message,
+					void *user_data)
+{
+	struct eap_secret_request *req = user_data;
+
+	req->network->agent_request = 0;
+	req->secret->value = l_strdup(value);
+
+	req->callback(result, message, req);
+}
+
+static void eap_user_password_callback(enum agent_result result,
+					const char *user, const char *passwd,
+					struct l_dbus_message *message,
+					void *user_data)
+{
+	struct eap_secret_request *req = user_data;
+
+	req->network->agent_request = 0;
+
+	if (user && passwd) {
+		size_t len1 = strlen(user) + 1;
+		size_t len2 = strlen(passwd) + 1;
+
+		req->secret->value = l_malloc(len1 + len2);
+		memcpy(req->secret->value, user, len1);
+		memcpy(req->secret->value + len1, passwd, len2);
+	}
+
+	req->callback(result, message, req);
+}
+
+static bool eap_send_agent_req(struct network *network,
+				struct l_queue *pending_secrets,
+				struct l_dbus_message *message,
+				void *callback)
+{
+	struct eap_secret_request *req;
+	struct eap_secret_info *info;
+
+	/*
+	 * Request the locally-verifiable data first, i.e.
+	 * the private key encryption passphrases so that we don't bother
+	 * asking for any other data if these passphrases turn out to
+	 * be wrong.
+	 */
+	info = l_queue_remove_if(pending_secrets, eap_secret_info_match_local,
+					NULL);
+
+	if (!info)
+		info = l_queue_pop_head(pending_secrets);
+
+	req = l_new(struct eap_secret_request, 1);
+	req->network = network;
+	req->secret = info;
+	req->pending_secrets = pending_secrets;
+	req->callback = callback;
+
+	switch (info->type) {
+	case EAP_SECRET_LOCAL_PKEY_PASSPHRASE:
+		network->agent_request = agent_request_pkey_passphrase(
+						network->object_path,
+						eap_password_callback,
+						message, req,
+						eap_secret_request_free);
+		break;
+	case EAP_SECRET_REMOTE_PASSWORD:
+		network->agent_request = agent_request_user_password(
+						network->object_path,
+						info->parameter,
+						eap_password_callback,
+						message, req,
+						eap_secret_request_free);
+		break;
+	case EAP_SECRET_REMOTE_USER_PASSWORD:
+		network->agent_request = agent_request_user_name_password(
+						network->object_path,
+						eap_user_password_callback,
+						message, req,
+						eap_secret_request_free);
+		break;
+	}
+
+	if (network->agent_request)
+		return true;
+
+	eap_secret_request_free(req);
+	return false;
+}
+
+static struct l_dbus_message *network_connect_8021x(struct network *network,
+						struct scan_bss *bss,
+						struct l_dbus_message *message);
+
+static void eap_secret_done(enum agent_result result,
+				struct l_dbus_message *message,
+				struct eap_secret_request *req)
+{
+	struct network *network = req->network;
+	struct eap_secret_info *secret = req->secret;
+	struct l_queue *pending = req->pending_secrets;
+	struct scan_bss *bss;
+
+	l_debug("result %d", result);
+
+	/*
+	 * Agent will release its reference to message after invoking this
+	 * callback.  So if we want this message, we need to take a reference
+	 * to it.
+	 */
+	l_dbus_message_ref(message);
+
+	if (result != AGENT_RESULT_OK) {
+		dbus_pending_reply(&message, dbus_error_aborted(message));
+		goto err;
+	}
+
+	bss = network_bss_select(network);
+
+	/* Did all good BSSes go away while we waited */
+	if (!bss) {
+		dbus_pending_reply(&message, dbus_error_failed(message));
+		goto err;
+	}
+
+	if (!network->secrets)
+		network->secrets = l_queue_new();
+
+	l_queue_push_tail(network->secrets, secret);
+
+	req->secret = NULL;
+
+	/*
+	 * If we have any other missing secrets in the queue, send the
+	 * next request immediately unless we've just received a passphrase
+	 * for a local private key.  In that case we will first call
+	 * network_connect_8021x to have it validate the new passphrase.
+	 */
+	if (secret->type == EAP_SECRET_LOCAL_PKEY_PASSPHRASE ||
+			l_queue_isempty(req->pending_secrets)) {
+		struct l_dbus_message *reply;
+
+		reply = network_connect_8021x(network, bss, message);
+		if (reply)
+			dbus_pending_reply(&message, reply);
+		else
+			l_dbus_message_unref(message);
+
+		return;
+	}
+
+	req->pending_secrets = NULL;
+
+	if (eap_send_agent_req(network, pending, message,
+				eap_secret_done)) {
+		l_dbus_message_unref(message);
+		return;
+	}
+
+	dbus_pending_reply(&message, dbus_error_no_agent(message));
+err:
+	network_settings_close(network);
+}
+
+static struct l_dbus_message *network_connect_8021x(struct network *network,
+						struct scan_bss *bss,
+						struct l_dbus_message *message)
+{
+	int r;
+	struct l_queue *missing_secrets = NULL;
+	struct l_dbus_message *reply;
+
+	l_debug("");
+
+	r = eap_check_settings(network->settings, network->secrets, "EAP-",
+				true, &missing_secrets);
+	if (r) {
+		if (r == -EUNATCH)
+			reply = dbus_error_not_available(message);
+		else if (r == -ENOTSUP)
+			reply = dbus_error_not_supported(message);
+		else if (r == -EACCES)
+			reply = dbus_error_failed(message);
+		else
+			reply = dbus_error_not_configured(message);
+
+		goto error;
+	}
+
+	l_debug("supplied %u secrets, %u more needed for EAP",
+		l_queue_length(network->secrets),
+		l_queue_length(missing_secrets));
+
+	if (l_queue_isempty(missing_secrets)) {
+		if (!network_set_8021x_secrets(network)) {
+			reply = dbus_error_failed(message);
+
+			goto error;
+		}
+
+		device_connect_network(network->device, network, bss, message);
+
+		return NULL;
+	}
+
+	if (eap_send_agent_req(network, missing_secrets, message,
+				eap_secret_done))
+		return NULL;
+
+	reply = dbus_error_no_agent(message);
+
+error:
+	network_settings_close(network);
+
+	l_queue_destroy(network->secrets, eap_secret_info_free);
+	network->secrets = NULL;
+
+	return reply;
+}
+
 static struct l_dbus_message *network_connect(struct l_dbus *dbus,
 						struct l_dbus_message *message,
 						void *user_data)
@@ -664,13 +977,59 @@ static struct l_dbus_message *network_connect(struct l_dbus *dbus,
 		if (!network_settings_load(network))
 			return dbus_error_not_configured(message);
 
-		device_connect_network(device, network, bss, message);
-		return NULL;
+		return network_connect_8021x(network, bss, message);
 	default:
 		return dbus_error_not_supported(message);
 	}
 }
 
+void network_connect_new_hidden_network(struct network *network,
+						struct l_dbus_message *message)
+{
+	struct device *device = network->device;
+	struct scan_bss *bss;
+	struct l_dbus_message *error;
+
+	l_debug("");
+
+	/*
+	 * This is not a Known Network.  If connection succeeds, either
+	 * network_sync_psk or network_connected will save this network
+	 * as hidden and trigger an update to the hidden networks count.
+	 */
+	network->info->is_hidden = true;
+
+	bss = network_bss_select(network);
+	if (!bss) {
+		/* This should never happened for the hidden networks. */
+		error = dbus_error_not_supported(message);
+		goto reply_error;
+	}
+
+	network->settings = l_settings_new();
+	l_settings_set_bool(network->settings, "Settings", "Hidden", true);
+
+	switch (network_get_security(network)) {
+	case SECURITY_PSK:
+		error = network_connect_psk(network, bss, message);
+		break;
+	case SECURITY_NONE:
+		device_connect_network(device, network, bss, message);
+		return;
+	default:
+		error = dbus_error_not_supported(message);
+		break;
+	}
+
+	if (error)
+		goto reply_error;
+
+	return;
+
+reply_error:
+	dbus_pending_reply(&message, error);
+}
+
 static bool network_property_get_name(struct l_dbus *dbus,
 					struct l_dbus_message *message,
 					struct l_dbus_message_builder *builder,
@@ -780,6 +1139,9 @@ void network_remove(struct network *network, int reason)
 	if (network->object_path)
 		network_unregister(network, reason);
 
+	l_queue_destroy(network->secrets, eap_secret_info_free);
+	network->secrets = NULL;
+
 	l_queue_destroy(network->bss_list, NULL);
 	network_info_put(network->info);
 
@@ -846,7 +1208,9 @@ void network_rank_update(struct network *network)
 			n = L_ARRAY_SIZE(rankmod_table) - 1;
 
 		rank = rankmod_table[n] * best_bss->rank + USHRT_MAX;
-	} else if (network->info->is_known)
+	} else if (!l_queue_find(networks, network_info_ptr_match,
+					network->info))
+		/* Is a known network */
 		rank = best_bss->rank;
 	else
 		rank = (int) best_bss->rank - USHRT_MAX; /* Negative rank */
@@ -854,27 +1218,24 @@ void network_rank_update(struct network *network)
 	network->rank = rank;
 }
 
-bool network_info_add_known(const char *ssid, enum security security)
+struct network_info *network_info_add_known(const char *ssid,
+						enum security security)
 {
 	struct network_info *network;
-	int err;
+	struct network_info search;
+
+	strcpy(search.ssid, ssid);
+	search.type = security;
+
+	network = l_queue_remove_if(networks, network_info_match, &search);
+	if (network)
+		return network;
 
 	network = l_new(struct network_info, 1);
 	strcpy(network->ssid, ssid);
 	network->type = security;
 
-	err = storage_network_get_mtime(security_to_str(security), ssid,
-					&network->connected_time);
-	if (err < 0) {
-		l_free(network);
-		return false;
-	}
-
-	network->is_known = true;
-
-	l_queue_insert(networks, network, timespec_compare, NULL);
-
-	return true;
+	return network;
 }
 
 static void network_info_check_device(struct device *device, void *user_data)
@@ -888,39 +1249,16 @@ static void network_info_check_device(struct device *device, void *user_data)
 		device_disconnect(device);
 }
 
-bool network_info_forget_known(const char *ssid, enum security security)
+void network_info_forget_known(struct network_info *network)
 {
-	struct network_info *network, search;
-
-	search.type = security;
-	strcpy(search.ssid, ssid);
-
-	network = l_queue_remove_if(networks, network_info_match, &search);
-	if (!network)
-		return false;
-
-	if (!network->seen_count) {
-		network_info_free(network);
-
-		return true;
-	}
-
-	memset(&network->connected_time, 0, sizeof(struct timespec));
-
-	network->is_known = false;
-
-	l_queue_push_tail(networks, network);
-
 	__iwd_device_foreach(network_info_check_device, network);
 
-	return true;
-}
-
-void network_info_foreach(network_info_foreach_func_t function,
-				void *user_data)
-{
-	const struct l_queue_entry *entry;
-
-	for (entry = l_queue_get_entries(networks); entry; entry = entry->next)
-		function(entry->data, user_data);
+	/*
+	 * Network is no longer a Known Network, see if we still need to
+	 * keep the network_info around.
+	 */
+	if (network->seen_count)
+		l_queue_push_tail(networks, network);
+	else
+		network_info_free(network);
 }
